/** @file
  Starting point for the core memory reference code.

@copyright
  INTEL CONFIDENTIAL
  Copyright 1999 - 2020 Intel Corporation.

  The source code contained or described herein and all documents related to the
  source code ("Material") are owned by Intel Corporation or its suppliers or
  licensors. Title to the Material remains with Intel Corporation or its suppliers
  and licensors. The Material may contain trade secrets and proprietary and
  confidential information of Intel Corporation and its suppliers and licensors,
  and is protected by worldwide copyright and trade secret laws and treaty
  provisions. No part of the Material may be used, copied, reproduced, modified,
  published, uploaded, posted, transmitted, distributed, or disclosed in any way
  without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure or delivery
  of the Materials, either expressly, by implication, inducement, estoppel or
  otherwise. Any license under such intellectual property rights must be
  express and approved by Intel in writing.

  Unless otherwise agreed by Intel in writing, you may not remove or alter
  this notice or any other notice embedded in Materials by Intel or
  Intel's suppliers or licensors in any way.

  This file contains an 'Intel Peripheral Driver' and is uniquely identified as
  "Intel Reference Module" and is licensed for Intel CPUs and chipsets under
  the terms of your license agreement with Intel or your vendor. This file may
  be modified by the user, subject to additional terms of the license agreement.

@par Specification Reference:
**/

//
// Include files
//

#include "MrcTypes.h"
#include "MrcApi.h"
#include "MrcCommandTraining.h"
#include "MrcCommon.h"
#include "MrcCrosser.h"
#include "MrcDdr3.h"
#include "MrcDebugHook.h"
#include "MrcDebugPrint.h"
#include "MrcGeneral.h"
#include "MrcGlobal.h"
#include "MrcBdat.h"
#include "MrcMcConfiguration.h"
#include "MrcMemoryMap.h"
#include "MrcMemoryScrub.h"
#include "MrcReadDqDqs.h"
#include "MrcReadReceiveEnable.h"
#include "MrcRegisterCache.h"
#include "MrcReset.h"
#include "MrcSaveRestore.h"
#include "MrcSpdProcessing.h"
#include "MrcMaintenance.h"
#include "MrcStartMemoryConfiguration.h"
#include "MrcWriteDqDqs.h"
#include "MrcWriteLeveling.h"
#include "MrcGears.h"
#include "MrcDdrIoApi.h"

/**
  Print the input parameters to the debug message output port.

  @param[in] MrcData - The MRC global data.

  @retval mrcSuccess
**/

//
// Functions:
//
const CallTableEntry  MrcCallTable[] = {
  ///
  /// The functions are executed in the following order, as the policy flag dictates.
  /// Mrctask, post_code, OEM command, policy_flag, iteration, debug_string
  ///
  {MrcEarlyOverrides,               MRC_EARLY_OVERRIDES,        OemEarlyOverrides,      1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST                           ,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Early Overrides")},
  {MrcSafeMode,                     MRC_SAFE_MODE,              OemMrcSafeMode,         1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST                           ,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Safe Mode Override")},
  {MrcFastBootPermitted,            MRC_FAST_BOOT_PERMITTED,    OemFastBootPermitted,   1,                             MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Fast boot permitted")},
  {MrcRestoreNonTrainingValues,     MRC_RESTORE_NON_TRAINING,   OemRestoreNonTraining,  1,           MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Restore non-training values")},
#ifdef MRC_DEBUG_PRINT
  {MrcPrintInputParameters,         MRC_PRINT_INPUT_PARAMS,     OemPrintInputParameters,1, MF_COLD | MF_WARM         | MF_FAST | MF_GV_FIRST                           ,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Print input parameters")},
#endif // MRC_DEBUG_PRINT
  {MrcSpdProcessing,                MRC_SPD_PROCESSING,         OemSpdProcessingRun,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("SPD PROCESSING")},
  {MrcSetOverrides,                 MRC_SET_OVERRIDES,          OemSetOverride,         1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Auto Updates and Overrides")},
  {MrcMcCapability,                 MRC_MC_CAPABILITY,          OemMcCapability,        1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MC Capabilities")},
  {MrcFrequencyLock,                MRC_FREQ_LOCK,              OemFrequencySet,        1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MC Frequency Lock")},
  {MrcMemorySsInit,                 MRC_MEMORY_SS_INIT,         OemMemorySsInit,        1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Memory SubSystem Config")},
  {MrcDdrIoPreInit,                 MRC_DDRIO_PREINIT,          OemDdrPhyInit,          1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DDRIO Config - Pre DCC")},
  {MrcMcConfiguration,              MRC_MC_CONFIG,              OemMcInitRun,           1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MC Config")},
  {MrcDdrIoInit,                    MRC_DDRIO_INIT,             OemDdrPhyInit,          1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DDRIO Config - Post DCC")},
  {MrcMbist,                        MRC_MBIST,                  OemMbist,               1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Memory Subsystem MBIST") },
  {MrcFllInit,                      MRC_FLL_INIT,               OemFllTraining,         1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("FLL Init")},
  {MrcDdrCompCalPre,                MRC_DDRCOMP_CAL_PRE,        OemDdrCompCalPre,       1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DDR COMP Calc - PreTraining")},
  {MrcDdrScomp,                     MRC_DDRSCOMP_INIT,          OemDdrScompInit,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DDR SCOMP Config")},
  {MrcSetMemoryMap,                 MRC_MC_MEMORY_MAP,          OemMcMemoryMap,         1, MF_COLD | MF_WARM | MF_S3 | MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MC MEMORY MAP")},
  {MrcPreTraining,                  MRC_PRE_TRAINING,           OemPreTraining,         1, MF_COLD | MF_WARM         | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Pre-Training")},
  {MrcResetRecovery,                MRC_RESET_RECOVERY,         OemResetRecovery,       1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Reset Recovery")},
  {MrcOffsetCorrectionPre,          MRC_OFFSET_CORRECTION_PRE,  OemOffsetCorrectionPre, 1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT ("DDR Offset Correction - PreTraining")},
  {MrcDccInitPreJedec,              MRC_DCC_INIT,               OemDccInit,             1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DCC Init - PreJedecReset")},

  // For LPDDR, we switch to low freq before doing JEDEC reset sequence and then switch to high freq after JEDEC reset sequence
  {MrcResetSequence,                MRC_RESET_SEQUENCE,         OemMcResetRun,          1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("JEDEC RESET")},
  {MrcPrintDdrMrs,                  MRC_PRINT_DDR_MRS,          OemPrintDdrMrs,         1, MF_COLD                   | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Print DDR MRs")},
  {MrcSenseAmpOffsetTraining,       MRC_SENSE_AMP_OFFSET,       OemSenseAmpTraining,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("SenseAmp Offset Training")},
  {MrcDccPreEct,                    MRC_DCC_PRE_ECT,            OemDccPreEct,           1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DCC - Pre ECT")},
  // Switch to low frequency
  {MrcEarlyCommandTraining,         MRC_EARLY_COMMAND,          OemEarlyCommandTraining,1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Early command training")},
  // We are at the target frequency
  {MrcDccPreLeveling,               MRC_DCC_PRE_LEVELING,       OemDccPreLeveling,      1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DCC Pre Leveling")},
  {MrcJedecWCKTraining,             MRC_WCK_LEVELING,           OemWckLeveling,         1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Jedec WCK training")},
  //{MrcEarlyReadMprTimingCentering2D,MRC_EARLY_RDMPR_TIMING_2D,  OemEarlyRmpr2DPostEct,  1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Early ReadMPR Timing Centering 2D High Freq")},
  {MrcPdaEnumeration,               MRC_DCC_PDA_ENUMERATE,      OemPdaEnumeration,      1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("PDA Enumeration")},
  {MrcReadLevelingTraining,         MRC_RECEIVE_ENABLE,         OemReceiveEnable,       1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Leveling training")},
  {MrcReadMprTraining,              MRC_READ_MPR,               OemReadMprTraining,     1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("CPGC-Based Read MPR training")},
  {MrcJedecWriteLevelingTraining,   MRC_JEDEC_WRITE_LEVELING,   OemJedecWriteLeveling,  1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Jedec Write Leveling training")},
  {MrcWriteLevelingFlyByTraining,   MRC_WRITE_LEVELING_FLYBY,   OemWriteLevelingFlyby,  1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Leveling FlyBy training")},
  {MrcDccPostWriteLeveling,         MRC_DCC_POST_WR_LEVELING,   OemDccPostWrLeveling,   1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DCC Post Write Leveling")},
  {MrcWeakWriteTimingCentering,     MRC_WEAK_WRITE_TIMING_1D,   OemWeakWriteDqDqs,      1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Weak Write Timing Centering before Early 2Ds")},
  {MrcEarlyWriteTimingCentering2D,  MRC_EARLY_WRITE_TIMING_2D,  OemEarlyWriteDqDqs2D,   1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Early Write Timing Centering 2D")},
  {MrcEarlyReadTimingCentering2D,   MRC_EARLY_READ_TIMING_2D,   OemEarlyReadDqDqs2D,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Early Read Timing Centering 2D")},
  {MrcWriteTimingCentering,         MRC_WRITE_TIMING_1D,        OemWriteDqDqs,          1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Timing Centering")},
//{MrcDdr4DeviceMap,                MRC_DDR4_MAPPING,           OemDdr4Map,             1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DDR4 PDA Mapping")},
  {MrcWriteVoltageCentering,        MRC_WRITE_VOLTAGE_1D,       OemWriteVoltage,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Voltage Centering")},
  {MrcReadVoltageCentering,         MRC_READ_VOLTAGE_1D,        OemReadVoltage,         1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT,          MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Voltage Centering")},
  {MrcReadTimingCentering,          MRC_READ_TIMING_1D,         OemReadDqDqs,           1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Timing Centering")},
  {MrcWriteTimingCentering2D,       MRC_WRITE_TIMING_2D,        OemWriteDqDqs2D,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Timing Centering 2D Before Power Training")},
  {MrcReadTimingCentering2D,        MRC_READ_TIMING_2D,         OemReadDqDqs2D,         1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Timing Centering 2D Before Power Training")},
  {MrcPowerSavingMeter,             MRC_PWR_MTR,                OemPowerSavingMeter,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("PowerSavingMeter Base Line Update")},
  {MrcTxDqTcoCompTraining,          MRC_TXDQTCO_COMP_TRAINING,  OemTxDqTcoCompTraining, 1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("TxDqTCO Comp Training")},
  {MrcTxDqsTcoCompTraining,         MRC_TXDQSTCO_COMP_TRAINING, OemTxDqsTcoCompTraining,1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("TxDqsTCO Comp Training")},
  {MrcCmdVoltageCentering,          MRC_CMD_VREF,               OemCmdVoltCentering,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Command Voltage Centering")},
// This is used for debug
//  {CommandPlot2D,                 MRC_CMD_PLOT_2D,            OemReadDqDqs,           1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("2D command plot")},
  {MrcPostTraining,                 MRC_POST_TRAINING,          OemPostTraining,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Post-training")},

//  {MrcVccDLLBypassTraining,         MRC_VCCDLL_BYPASS,          OemVccDllBypassTraining,1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("VccDLL bypass to VccIOG training")},
  {MrcLateCommandTraining,          MRC_LATE_COMMAND,           OemLateCommandTraining, 1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Late command training")},
  {MrcCmdVoltageCentering,          MRC_CMD_VREF,               OemCmdVoltCentering,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Command Voltage Centering")},
  {MrcDimmOdtCaTraining,            MRC_DIMM_ODT_CA,            OemDimmOdtCaTraining,   1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DIMM CA ODT Training")},
  {MrcCmdEqDsTraining,              MRC_CMD_EQ,                 OemCmdDriveStrengthEq,  1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("CMD/CTL Drive Strength and Equalization 2D")},
  {MrcCmdDsUpDnTraining,            MRC_CMD_DS,                 OemCmdDriveUpDn,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("CMD/CTL Drive Strength Up/Dn 2D")},
  {MrcCmdSlewRate,                  MRC_CMD_SR,                 OemCmdSlewRate,         1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("CMD CTL CLK Slew Rate") },
  {MrcDimmOdtTraining,              MRC_DIMM_ODT,               OemDimmODTTraining,     1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DIMM Wr ODT Training")},
  {MrcWriteEqDsTraining,            MRC_WRITE_EQ,               OemWriteDriveStrengthEq,1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Drive Strength and Equalization 2D")},
  {MrcWriteDsUpDnTraining,          MRC_WRITE_DS,               OemWriteDriveUpDn,      1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Drive Strength Up/Dn 2D")},
  {MrcWriteSlewRate,                MRC_WRITE_SR,               OemWriteSlewRate,       1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Slew Rate")},
//  {MrcPanicVttDnLpTraining,         MRC_PANIC_VTT_DN_LP,        OemPanicVttDnLpTraining,1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("PanicVttDnLp Training") },
  {MrcRxVrefDecapTraining,          MRC_READ_VREF_DECAP,        OemReadVrefDecapTraining,1, MF_COLD                            | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Vref Decap Training")},
  {MrcVddqTraining,                 MRC_VDDQ,                   OemVddqTraining,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Vddq Training") },
  {MrcReadAmplifierPower,           MRC_READ_AMP_POWER,         OemReadAmplifierPower,  1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Amplifier Power")},
  {MrcDimmRonTraining,              MRC_DIMM_RON,               OemDimmRonTraining,     1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DIMM Ron Training")},
  {MrcReadODTTraining,              MRC_READ_ODT,               OemReadODTTraining,     1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read ODT Training")},
  {MrcReadEQTraining,               MRC_READ_EQ,                OemReadEQTraining,      1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Equalization Training")},
  {MrcOptimizeComp,                 MRC_CMP_OPT,                OemOptimizeComp,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Comp Optimization Training")},
//  {MrcForceOltm,                    MRC_FORCE_OLTM,             OemForceOltm,           1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Force OLTM")},

//  {MrcPerBitDeskewCalibration,      MRC_PBD_DESKEW_CAL,         OemPerBitDeskewCal,     1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Per Bit Deskew Calibration")},
  {MrcWriteVoltageCentering2D,      MRC_WRITE_VREF_2D,          OemWriteVoltCentering2D,1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Voltage Centering 2D")},
  {MrcReadVoltageCentering2D,       MRC_READ_VREF_2D,           OemReadVoltCentering2D, 1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Voltage Centering 2D")},
  {MrcWriteTimingCentering2D,       MRC_WRITE_TIMING_2D,        OemWriteDqDqs2D,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write Timing Centering 2D After Power Training")},
  {MrcReadTimingCentering2D,        MRC_READ_TIMING_2D,         OemReadDqDqs2D,         1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Read Timing Centering 2D After Power Training")},
  {MrcReceiveEnTimingCentering,     MRC_RCVEN_TIMING_1D,        OemRcvEnCentering1D,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Receive Enable Timing Centering")},

//  {MrcRoundTripLatency,             MRC_ROUND_TRIP_LAT,         OemRoundTripLatency,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Round Trip Latency Training")},

//  {MrcVccDllFFTiming,               MRC_VCCDLL_FFTIMING,        OemVccDllFFTiming            ,1, MF_COLD                       | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("VccDll FF Timing")},
//  {MrcVccDllFFOffsetCorrection,     MRC_VCCDLL_FFOFFSETCORRECTION,OemVccDllFFOffsetCorrection,1, MF_COLD                       | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("VccDll FF Offset Correction")},
//  {MrcKickBackNoiseCorrection,      MRC_VCCDLL_KBNCORRECTION,   OemVccDllKBNCorrection       ,1, MF_COLD                       | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("VccDll KickBack Noise Correction")},
//  {MrcVsxHiFFTuning,                MRC_VSXHI_FFTUNING,         OemVsxhiFFTuning             ,1, MF_COLD                       | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("VccDllVsxHi Feed Forward tuning")},

  {MrcTurnAroundTiming,             MRC_TURN_AROUND,            OemTurnAroundTimes,     1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST | MF_RMT | MF_TST, MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Turn Around Trip Training")},

  {MrcRetrainMarginCheck,           MRC_RETRAIN_CHECK,          OemRetrainMarginCheck,  1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   1,                 MRC_DEBUG_TEXT("Check Margin for Retrain")},
  {MrcRmtEvenOdd,                   MRC_RMT_TOOL_EVENODD,       OemRmtEvenOdd,          1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Rank Margin Tool for Even Odd UI Errors")},
  {MrcRankMarginTool,               MRC_RMT_TOOL,               OemRmt,                 1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Rank Margin Tool")},
  {MrcRankMarginToolBit,            MRC_RMT_TOOL_BIT,           OemRmtPerBit,           1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Rank Margin Tool Bit")},
  {MrcPowerSavingMeter,             MRC_PWR_MTR,                OemPowerSavingMeter,    1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("PowerSavingMeter update")},
  {MrcWriteDqDqsReTraining,         MRC_WRITE_DQ_DQS_RETRAINING,OemWriteDqDqsReTraining,1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Write DQ/DQS Retraining")},
  {MrcSaGvFinal,                    MRC_MR_FILL,                OemNumOfCommands,       1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("SAGV Finalization")},
  //@todo {MrcDccReplica,                   MRC_DCC_REPLICA,            OemDccReplica,          1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DCC Replica")},
  //@todo {MrcDccFinalization,              MRC_DCC_FINAL,              OemDccFinal,            1, MF_COLD                                                     | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DCC Finalization for Cold fow")},
  {MrcMcActivate,                   MRC_MC_ACTIVATE,            OemMrcActivate,         1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC activate")},
  {MrcIbecc,                        MRC_IBECC,                  OemIbecc,               1, MF_COLD | MF_WARM         | MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT ("MRC In-band ECC") },
  {MrcAliasCheck,                   MRC_ALIAS_CHECK,            OemAliasCheck,          1, MF_COLD                                                         | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Memory alias check")},
  {MrcEccClean,                     MRC_ECC_CLEAN_START,        OemHwMemInit,           1, MF_COLD                                                         | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Memory Scrubbing")},
  {MrcNormalMode,                   MRC_NORMAL_MODE,            OemNormalMode,          1, MF_COLD                                                         | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Normal Operation")},
  {MrcSaveMCValues,                 MRC_SAVE_MC_VALUES,         OemSaveMCValues,        1, MF_COLD                             | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Save MC Values")},
  {MrcUpdateSavedMCValues,          MRC_UPDATE_SAVE_MC_VALUES,  OemUpdateSaveMCValues,  1,                             MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Update MC Values in Fast flow")},
  {MrcRestoreTrainingValues,        MRC_RESTORE_TRAINING,       OemRestoreTraining,     1,           MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Restore Training Values")},
  //@todo {MrcDccFinalization,              MRC_DCC_FINAL,              OemDccFinal,            1,           MF_WARM | MF_S3 | MF_FAST                         | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("DCC Finalization for non-Cold flows")},
  {MrcPdaEnumeration,               MRC_DCC_PDA_ENUMERATE,      OemPdaEnumeration,      1, MF_WARM |MF_FAST                    | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("PDA Enumeration")},
  {MrcResetSequence,                MRC_RESET_SEQUENCE,         OemMcResetRun,          1,                             MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("JEDEC RESET on Fast flow")},
  {MrcEccClean,                     MRC_ECC_CLEAN_START,        OemHwMemInit,           1,                             MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Memory Scrubbing")},
//{MrcRankMarginTool,               MRC_RMT_TOOL,               OemRmt,                 1,                             MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Rank Margin Tool on Fast flow")},
  {MrcNormalMode,                   MRC_NORMAL_MODE,            OemNormalMode,          1,                             MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Normal Operation For Fast flow") },
  {MrcSelfRefreshExit,              MRC_SELF_REFRESH_EXIT,      OemSelfRefreshExit,     1,           MF_WARM | MF_S3                                       | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Perform Self Refresh Exit")},
  {MrcEccClean,                     MRC_ECC_CLEAN_START,        OemHwMemInit,           1,           MF_WARM                                               | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Memory Scrubbing")},
  {MrcNormalMode,                   MRC_NORMAL_MODE,            OemNormalMode,          1,           MF_WARM | MF_S3                                       | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Normal Operation For Warm / S3 flow")},
  {MrcThermalConfig,                MRC_THERMAL_CONFIG,         OemThermalConfig,       1, MF_COLD | MF_WARM | MF_S3 | MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("Configure Thermal values")},
  {MrcEnergyPerfGain,               MRC_ENG_PERF_GAIN,          OemEngPerfGain,         1, MF_COLD | MF_WARM | MF_S3 | MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Energy Performance Gain")},
  {MrcSaGvSwitch,                   MRC_SA_GV_SWITCH,           OemSaGvSwitch,          1, MF_COLD | MF_WARM | MF_S3 | MF_FAST | MF_GV_FIRST | MF_GV_OTHER | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC SA GV Switch")},
  {MrcTxtAliasCheck,                MRC_TXT_ALIAS_CHECK,        OemTxtAliasCheck,       1, MF_COLD | MF_WARM         | MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("TXT Memory alias check")},
  {MrcDone,                         MRC_DONE,                   OemMrcDone,             1, MF_COLD | MF_WARM | MF_S3 | MF_FAST                             | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC done")},
#ifdef BDAT_SUPPORT
  {MrcFillBdatStructure,            MRC_FILL_BDAT_STRUCTURE,    OemMrcFillBdat,         1, MF_COLD                                                         | MF_GV_LAST,                   MRC_ITERATION_MAX, MRC_DEBUG_TEXT("MRC Fill BDAT Structure")},
#endif // BDAT_SUPPORT
};

/**
  this function use by the OEM to do dedicated task during the MRC.

  @param[in] MrcData           - include all the MRC data
  @param[in] OemStatusCommand  - A command that indicates the task to perform.
  @param[in] ptr               - general ptr for general use.

  @retval The status of the task.
**/
MrcStatus
MrcInternalCheckPoint (
  IN MrcParameters        *MrcData,
  IN MrcOemStatusCommand  OemStatusCommand,
  IN void                 *ptr
  )
{
  MrcInput           *Inputs;
  MRC_FUNCTION       *MrcCall;
  MrcOutput          *Outputs;
  MrcStatus          Status;
  MrcDdrType         DdrType;
  TrainingStepsEn    *TrainingEnables;
  TrainingStepsEn2   *TrainingEnables2;
  BOOLEAN            Ddr4;
  BOOLEAN            Ddr5;
  BOOLEAN            Lpddr4;
  BOOLEAN            Lpddr5;
  BOOLEAN            Lpddr;

  Status      = mrcSuccess;
  Inputs      = &MrcData->Inputs;
  MrcCall     = Inputs->Call.Func;
  Outputs     = &MrcData->Outputs;
  DdrType     = Outputs->DdrType;
  TrainingEnables   = &Inputs->TrainingEnables;
  TrainingEnables2  = &Inputs->TrainingEnables2;

  Lpddr4 = (DdrType == MRC_DDR_TYPE_LPDDR4);
  Lpddr5 = (DdrType == MRC_DDR_TYPE_LPDDR5);
  Ddr4   = (DdrType == MRC_DDR_TYPE_DDR4);
  Ddr5   = (DdrType == MRC_DDR_TYPE_DDR5);
  Lpddr  = (Lpddr4 || Lpddr5);

  switch (OemStatusCommand) {
    case OemMrcSafeMode:
      if (!Inputs->MrcSafeConfig) {
        Status = mrcFail; // Skip this step
      }
      break;

    case OemSpdProcessingRun:
      break;

    case OemPreTraining:
      break;

    case OemResetRecovery:
    case OemOffsetCorrectionPre:
    case OemFllTraining:
    case OemDdrCompCalPre:
    case OemMcResetRun:
    case OemOptimizeComp:
    case OemPowerSavingMeter:
      if (Inputs->SimicsFlag == 1) {
        Status = mrcFail; // Don't run on SIMICS.
      }
      break;

    case OemTxDqTcoCompTraining:
      if (!TrainingEnables2->TXTCO) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemTxDqsTcoCompTraining:
      if (!TrainingEnables2->TXTCODQS) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemClkTcoCompTraining:
      if (!TrainingEnables2->CLKTCO) {
        Status = mrcFail; // Skip this step
      }
      break;

    case OemEarlyCommandTraining:
      if ((!TrainingEnables->ECT) || Ddr4) { // Cannot run ECT on DDR4 due to PDA not stable before basic training
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemPdaEnumeration:
      if (Inputs->SimicsFlag == 0) {
        if (!TrainingEnables2->PDA || !Ddr5) { // Only Run on Ddr5
          Status = mrcFail; // Skip this training step
        }
      } else {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemSenseAmpTraining:
      if (!TrainingEnables->SOT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadMprTraining:
      if (!TrainingEnables->RDMPRT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReceiveEnable:
      if (!TrainingEnables->RCVET) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWckLeveling:
      if (Lpddr5) {
        if (!TrainingEnables->JWRL) {
          Status = mrcFail; // Skip this training step
        }
      } else {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemJedecWriteLeveling:
      if (!Lpddr5) {
        if (!TrainingEnables->JWRL) {
          Status = mrcFail; // Skip this training step
        }
      } else {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteLevelingFlyby:
      if (!(Lpddr5 || Ddr5)) {
        if (!TrainingEnables->JWRL) {
          Status = mrcFail; // Skip this training step
        }
      } else {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteDqDqs:
      if (Inputs->SimicsFlag == 0) {
        if (!TrainingEnables->WRTC1D) {
          Status = mrcFail; // Skip this training step
        }
      } else {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWeakWriteDqDqs:
      if (Inputs->SimicsFlag == 0) {
        if (!TrainingEnables->WRTC1D || !Lpddr5) { // This step is only run for LP5
          Status = mrcFail; // Skip this training step
        }
      } else {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteDqDqsReTraining:
      if (!Lpddr || !Inputs->LpDqsOscEn || Lpddr5) { // This is for LPDDR4 only // WA is to disable it for LP5 while debug
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadVoltage:
      if (!TrainingEnables2->RDVC1D) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteVoltage:
      if (!TrainingEnables->WRVC1D) {
          Status = mrcFail; // Skip this training step
        }
      break;

    case OemEarlyWriteDqDqs2D:
      if (!TrainingEnables->EWRTC2D) { // needed for DDR4, Lp4 and Lp5
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemEarlyReadDqDqs2D:
      if (!TrainingEnables->ERDTC2D) { // This is for DDR4, Lp4 and Lp5
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemEarlyRmpr2DPostEct:
      if (!Lpddr) {
        Status = mrcFail;
        break;
      }
      /* FALLTHROUGH*/
      // Fall through if we are LPDDR4 to check input parameters.
    case OemEarlyReadMprDqDqs2D:
      if (!TrainingEnables->ERDMPRTC2D) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadDqDqs:
      if (!TrainingEnables->RDTC1D) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemPerBitDeskewCal:
      if (Inputs->RdTPbdDis && Inputs->WrTPbdDis) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemDdr4Map:
      if ((!Inputs->DDR4MAP) || !Ddr4) { // This is for DDR4 only
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemDimmODTTraining:
      if ((!TrainingEnables->DIMMODTT)) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemDimmOdtCaTraining:
      if ((!TrainingEnables2->DIMMODTCA) || !Lpddr) {  // This is for Lpddr only
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemDimmRonTraining:
      if ((!TrainingEnables->DIMMRONT)) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteSlewRate:
      if ((!TrainingEnables->WRSRT) || Lpddr) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteDriveUpDn:
      if (!TrainingEnables->WRDSUDT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteDriveStrengthEq:
      if (!TrainingEnables->WRDSEQT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadODTTraining:
      if (!TrainingEnables->RDODTT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadEQTraining:
      if (!TrainingEnables->RDEQT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemPostTraining:
      if ((!Ddr4) || !TrainingEnables->LCT) { // Cannot switch to 1N/2N without running LCT
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadAmplifierPower:
      if (!TrainingEnables->RDAPT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteDqDqs2D:
        if (!TrainingEnables->WRTC2D) {
          Status = mrcFail; // Skip this training step
        }
      break;

    case OemReadDqDqs2D:
      if (!TrainingEnables->RDTC2D) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemCmdVoltCentering:
      if (!TrainingEnables->CMDVC) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemCmdSlewRate:
      if ((!TrainingEnables2->CMDSR) || Lpddr){
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemCmdDriveStrengthEq:
      if (!TrainingEnables2->CMDDSEQ) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemWriteVoltCentering2D:
      if (!TrainingEnables->WRVC2D) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadVoltCentering2D:
      if (!TrainingEnables->RDVC2D) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemLateCommandTraining:
      if (!TrainingEnables->LCT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemRoundTripLatency:
      if ((!TrainingEnables->RTL) || Lpddr4 || Ddr4) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemTurnAroundTimes:
      if ((!TrainingEnables->TAT) || Lpddr4) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemRcvEnCentering1D:
      if ((!TrainingEnables->RCVENC1D) || Ddr4) {
        Status = mrcFail;  // Skip this step for non-LPDDR
      }
      break;

    case OemRetrainMarginCheck:
      if ((!TrainingEnables->RMC) || (Inputs->PowerTrainingMode != MrcTmPower)) {
        Status = mrcFail; // Skip if disabled or not in Power Training mode
      }
      break;


    case OemRmt:
      if (!TrainingEnables->RMT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemRmtEvenOdd:
      if (!TrainingEnables->RMT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemRmtPerBit:
      if (!TrainingEnables2->RMTBIT) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemMemTest:
      if ((!TrainingEnables->MEMTST) || (Inputs->SimicsFlag == 1)) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemAliasCheck:
      if (!TrainingEnables->ALIASCHK) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemBeforeNormalMode:
      break;

    case OemAfterNormalMode:
      break;

    case OemThermalConfig:
      break;

    case OemFrequencySetDone:
      break;

    case OemSaGvSwitch:
      if ((Inputs->SaGv != MrcSaGvEnabled) && (!Lpddr4)) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemDccPreLeveling:
    case OemDccPreReadTiming:
    case OemDccPostWrLeveling:
      if (Inputs->NewFeatureEnable1 == 0) {
        Status = mrcFail; //Skip this step
        break;
      }
      /* FALLTHROUGH*/
    case OemDccInit:
    case OemDccPreEct:
      if (!(TrainingEnables2->DCC && Outputs->Gear2)) {
        Status = mrcFail; // Skip this step
      }
      break;

    case OemVccDllBypassTraining:
      if ((!TrainingEnables2->VCCDLLBP) || (Inputs->PowerTrainingMode != MrcTmPower)) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemCmdDriveUpDn:
      if (!TrainingEnables2->CMDDRUD) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemPanicVttDnLpTraining:
      if ((!TrainingEnables2->PVTTDNLP) || (!Lpddr)) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemReadVrefDecapTraining:
      if (!TrainingEnables2->RDVREFDC) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemVddqTraining:
      if ((!TrainingEnables2->VDDQT) || (Inputs->PowerTrainingMode != MrcTmPower)) {
        Status = mrcFail; // Skip this training step
      }
      break;

    case OemHwMemInit:
      if (Inputs->SimicsFlag == 1) {
        Status = mrcFail; // Don't run on SIMICS.
      }
      break;

    default:
      break;
  }
  //
  // Allow Overriding the internal checkpoint decision.
  //
  MrcCall->MrcCheckpoint (MrcData, OemStatusCommand, &Status);
  return Status;
}

/**
  Prepare MRC structures for next MRC Iteraion (SAGV or margin retrain).

  @param[in, out] MrcData - Include all MRC global data.
  @param[in]      Retrain - Turn on/off UpmPwrRetrainFlag.

**/
void
MrcPrepareNextMrcIteration (
  IN OUT MrcParameters *const MrcData,
  IN BOOLEAN                  Retrain
  )
{
  UINT16                  PrevRdDataValid;
  BOOLEAN                 PrevGear2Ever;
  const MRC_FUNCTION      *MrcCall;
  MrcIntOutput            *MrcIntData;
  MrcOutput               *Outputs;
  MrcInput                *Inputs;
#ifdef BDAT_SUPPORT
  UINT16                    Index;
  MRC_BDAT_SCHEMA_LIST_HOB  *SaveBdatSchemasHob;
  BDAT_MEMORY_DATA_HOB      *SaveBdatMemoryHob[MAX_SCHEMA_LIST_LENGTH];
#endif
  Inputs  = &MrcData->Inputs;
  Outputs = &MrcData->Outputs;
  MrcIntData = ((MrcIntOutput *)(MrcData->IntOutputs.Internal));
  MrcCall = MrcData->Inputs.Call.Func;

  PrevRdDataValid = MrcIntData->MaxRdDataValid;
  PrevGear2Ever = Outputs->Gear2Ever;
#ifdef BDAT_SUPPORT
  SaveBdatSchemasHob = Outputs->BdatSchemasHob.Pointer;
  for (Index = 0; Index < MAX_SCHEMA_LIST_LENGTH; Index++) {
    SaveBdatMemoryHob[Index] = Outputs->BdatMemoryHob[Index].Pointer;
  }
#endif // BDAT_SUPPORT
  MrcCall->MrcSetMem ((UINT8 *) Outputs, sizeof (MrcOutput), 0);
  MrcCall->MrcSetMem ((UINT8 *) MrcIntData, sizeof (MrcIntOutput), 0);
  MrcIntData->Size = sizeof (MrcOutput);
  Outputs->Size = sizeof (MrcOutput);
  MrcIntData->MaxRdDataValid = PrevRdDataValid;
  Outputs->Gear2Ever = PrevGear2Ever;
  Outputs->UpmPwrRetrainFlag = Retrain;
#ifdef BDAT_SUPPORT
  Outputs->BdatSchemasHob.Pointer = SaveBdatSchemasHob;
  for (Index = 0; Index < MAX_SCHEMA_LIST_LENGTH; Index++) {
    Outputs->BdatMemoryHob[Index].Pointer = SaveBdatMemoryHob[Index];
  }
#endif // BDAT_SUPPORT

  MRC_DEBUG_MSG_OPEN (MrcData, Inputs->DebugLevel, Inputs->DebugStream);
}

/**
  Initializes the memory controller and DIMMs.

  @param[in, out] MrcData                - Include all MRC global data.
  @param[in]      Select                 - The post code of the call table entry to execute.
  @param[in, out] CurrentSaGvPoint       - The SAGV point that we want to execute the Select post code entry or to continue from (relevant if SAGV enabled).
  @param[in]      ExecuteFromSelectPoint - Execute Memory configuration from (Select,CurrentSaGvPoint) to the end

  @retval mrcSuccess if the initialization succeeded, otherwise an error status indicating the failure.
**/
MrcStatus
MrcStartMemoryConfiguration (
  IN OUT MrcParameters *const MrcData,
  IN     MrcPostCode          Select,
  IN OUT MrcSaGvPoint  *const CurrentSaGvPoint,
  IN BOOLEAN                  ExecuteFromSelectPoint
  )
{
  const MRC_FUNCTION      *MrcCall;
  const CallTableEntry    *Task;
  MrcDebug                *Debug;
  MrcInput                *Inputs;
  MrcOutput               *Outputs;
  MrcIntOutput            *MrcIntData;
  MrcSaveData             *SaveData;
  MrcSaGvPoint            SaGvPoint;
  MrcSaGvPoint            SaGvStart;
  MrcSaGvPoint            SaGvEnd;
  MrcSaGvPoint            SaGvFirst;
  MrcSaGvPoint            SaGvLast;
  MrcSaGv                 SaGv;
  MrcPostCode             post_code;
  MrcPostCode             PostCodeOut;
  MrcStatus               CpStatus;
  MrcStatus               MrcStatus;
  MRC_BOOT_MODE           BootMode;
  MrcUpmPwrRetrainLimits  RetrainLimits[MRC_NUMBER_UPM_PWR_RETRAIN_MARGINS];
  MrcFrequency            SaGvFreqToUse[MAX_SAGV_POINTS];
  UINT64                  start_time;
  UINT64                  finish_time;
  UINTN                   LowestStackAddr;
  INT32                   DebugLevel;
  UINT32                  ElapsedTime;
  UINT32                  TotalTime;
  UINT32                  PointTime;
  UINT32                  RetrainLoop;
  UINT16                  index;
  UINT8                   Run;
  UINT8                   RetrainDone;
  UINT8                   SaGvMask;
  UINT8                   SaGvBit;
  UINT8                   SaGvTracking;
  UINT8                   SaGvOther;
  UINT8                   SaGvGearToUse[MAX_SAGV_POINTS];
  BOOLEAN                 SaGvSpdFetched;

  Inputs  = &MrcData->Inputs;
  MrcCall = Inputs->Call.Func;
  Outputs = &MrcData->Outputs;
  Debug   = &Outputs->Debug;
  MrcIntData = ((MrcIntOutput *) (MrcData->IntOutputs.Internal));
  SaveData   = &MrcData->Save.Data;

  BootMode  = Inputs->BootMode;
  DebugLevel = Inputs->DebugLevel;

  // Initialize SaGv Point Optimization Tracking
  SaGvTracking = 0xFF;                  // Set to all high, 1111 1111 (Train all points)
  SaGvFirst = SaGvLast = MrcSaGvPoint0; // Set to first point; First and last SaGv Point to be trained (Associated with MF_SAGV_FIRST and MF_SAGV_LAST)
  SaGvOther = 0x0;                      // Set to all low, 0000 0000; track all points that aren't first or last (Associated with MF_SAGV_OTHER)
  SaGvSpdFetched = FALSE;

  //Initialize the Freq and gears for SaGv point from policy inputs
  for (int i = 0; i < MAX_SAGV_POINTS; i++) {
    SaGvFreqToUse[i] = (MrcFrequency)Inputs->SaGvFreq[i];
    SaGvGearToUse[i] = (UINT8)Inputs->SaGvGear[i];
  }

  if (Select == 0) {
    MrcVersionGet (MrcData, &Outputs->Version);
    MrcVersionPrint (MrcData, &Outputs->Version);
  }
  MrcStatus = (Select == 0) ? mrcFail : mrcSuccess;
  post_code = MRC_INITIALIZATION_START;
  Run       = 1;
  TotalTime = 0;
  LowestStackAddr = Debug->LowestStackAddr.DataN;

  MrcMcEarlyCapabilityCheck (MrcData);
  SaGv = Inputs->SaGv;

  if (SaGv == MrcSaGvEnabled) {
    SaGvStart = MrcSaGvPoint0;
    SaGvEnd = MrcSaGvPoint3;

    // Prep SaGv Optimization if point is given as auto (aka 0)
    for (SaGvPoint = MrcSaGvPoint0; SaGvPoint < MAX_SAGV_POINTS; SaGvPoint++) {
      // If and SaGv Point is set to auto, Need to know DdrType that is assigned to Outputs->DdrType during MrcSpdProcessing
      if ((BootMode == bmCold) && (!SaGvSpdFetched)) {
        MRC_DEBUG_MSG (Debug, MSG_LEVEL_NOTE, "\nMRC SaGv Optimization -- MrcSpdProcessing -- Started.\n");

        start_time  = MrcCall->MrcGetCpuTime ();
        MrcStatus = MrcSpdProcessing (MrcData);
        finish_time = MrcCall->MrcGetCpuTime ();

        SaGvSpdFetched = TRUE;

        MRC_DEBUG_MSG (
          Debug,
          MSG_LEVEL_ERROR,
          "\nMRC SaGv Optimization -- MrcSpdProcessing -- %s, Status = %Xh.\n",
          (mrcSuccess == MrcStatus) ? "SUCCEEDED" : "FAILED",
          MrcStatus
          );

        // Calculate elpased time for the this task and the entire MRC
        ElapsedTime = (UINT32) (finish_time - start_time);
        TotalTime += ElapsedTime;
        // <MrcTimer Task="Print input parameters" msec="2"/>
        MRC_DEBUG_MSG (Debug, MSG_LEVEL_TIME, "<MrcTimer Task=\"MrcSpdProcessing\" msec=\"%u\"/>\n", ElapsedTime);

        // Failed to fetch SPD data, we cannot optimize using the correct SaGv Freq and Gear
        if (MrcStatus != mrcSuccess) {
          return MrcStatus;
        }
      }

      // Correctly select the SaGv Freq and Gear
      MrcGetSagvConfig (MrcData, SaGvPoint, &SaGvFreqToUse[SaGvPoint], &SaGvGearToUse[SaGvPoint]);
    }

    // SaGvTracking is a bitmask initiallized to 1111 1111 where 1 mean must train this point and 0 mean we can skip to the HW handshake
    // Caveat: Points must be in order from lowest to highest.
    for (SaGvPoint = MAX_SAGV_POINTS - 1; SaGvPoint > 0; SaGvPoint--) {
      SaGvMask = 1 << SaGvPoint;
      SaGvBit = 1;

      if ((SaGvFreqToUse[SaGvPoint] == SaGvFreqToUse[SaGvPoint - 1]) && (SaGvGearToUse[SaGvPoint] == SaGvGearToUse[SaGvPoint - 1])) {
        SaGvBit = 0;
      }

      // Determine the point that should be associated with the MF_SAGV_LAST flag
      if ((SaGvLast == MrcSaGvPoint0) && (SaGvBit == 1)) {
        SaGvLast = SaGvPoint;
      }

      // Change the bit at position SaGvPoint to the value of SaGvBit, which is 1 (must train) or 0 (skip to HW handshake)
      SaGvTracking = (SaGvTracking & ~SaGvMask) | ((SaGvBit << SaGvPoint) & SaGvMask);
    } // for SaGvPoint

    MRC_DEBUG_MSG (Debug, MSG_LEVEL_NOTE, "\nMRC SaGv Optimization BitMask: 0x%x\n", SaGvTracking);

    // Determine the point that should be associated with MF_SAGV_LAST the MF_SAGV_OTHER Flag
    for (SaGvPoint = MAX_SAGV_POINTS - 1; SaGvPoint > 0; SaGvPoint--) {
      SaGvMask = 1 << SaGvPoint;
      if ((SaGvPoint != SaGvFirst) && (SaGvPoint != SaGvLast)) {
        SaGvOther |= SaGvMask;
      }
    } // for SaGvPoint

  } else {
    switch (SaGv) {
      case MrcSaGvFixed0:
        SaGvStart = MrcSaGvPoint0;
        break;

      case MrcSaGvFixed1:
        SaGvStart = MrcSaGvPoint1;
        break;

      case MrcSaGvFixed2:
        SaGvStart = MrcSaGvPoint2;
        break;

      case MrcSaGvDisabled:
      case MrcSaGvFixed3:
      default:
        SaGvStart = MrcSaGvPoint3;
        break;
    }
    SaGvEnd = SaGvLast = SaGvFirst = SaGvStart;
  }

  // Save the last point in the MRC save data
  SaveData->SaGvLast = SaGvLast;

  for (SaGvPoint = SaGvStart; Run && (SaGvPoint <= SaGvEnd); SaGvPoint++) {
    MrcIntData->SaGvPoint = SaGvPoint;
    // Update  *CurrentSaGvPoint for future calling
    if (CurrentSaGvPoint != NULL) {
      *CurrentSaGvPoint = SaGvPoint;
    }
    Inputs->Iteration = 0;
    PointTime = 0;

    // Reset UPM/PWR limits to initial values; previous SAGV point might have done retrain.
    MrcCall->MrcCopyMem (
      (UINT8 *) RetrainLimits,
      (UINT8 *) InitialLimits,
      sizeof (MrcUpmPwrRetrainLimits) * MRC_NUMBER_UPM_PWR_RETRAIN_MARGINS
      );

    MRC_DEBUG_MSG (
      Debug,
      MSG_LEVEL_NOTE,
      "SAGV: %sabled, running at point %u\n",
      (SaGv == MrcSaGvEnabled) ? "En" : "Dis",
      SaGvPoint
      );
    RetrainDone = 0;

    for (RetrainLoop = 0; RetrainLoop <= 1; RetrainLoop++) {
      MrcVersionGet (MrcData, &Outputs->Version);
      MrcIntData->UpmPwrRetrainLimits.Pointer = RetrainLimits;

      // If SaGv Point has already been trained, then skip training and perform hardware handshake.
      if ((SaGv == MrcSaGvEnabled) && (((SaGvTracking >> SaGvPoint) & 1) == 0)) {

        MRC_DEBUG_MSG (Debug, MSG_LEVEL_NOTE, "\nMRC SaGv Optimization -- MrcSaGvSwitch -- Started.\n");

        start_time  = MrcCall->MrcGetCpuTime ();
        MrcStatus = MrcSaGvSwitch (MrcData);
        finish_time = MrcCall->MrcGetCpuTime ();

        MRC_DEBUG_MSG (
          Debug,
          MSG_LEVEL_ERROR,
          "\nMRC SaGv Optimization -- MrcSaGvSwitch -- %s, Status = %Xh.\n",
          (mrcSuccess == MrcStatus) ? "SUCCEEDED" : "FAILED",
          MrcStatus
          );

        // Calculate elpased time for the this task and the entire MRC
        ElapsedTime = (UINT32) (finish_time - start_time);
        TotalTime += ElapsedTime;
        PointTime += ElapsedTime;
        // <MrcTimer Task="Print input parameters" msec="2"/>
        MRC_DEBUG_MSG (Debug, MSG_LEVEL_TIME, "<MrcTimer Task=\"MrcSaGvSwitch\" msec=\"%u\"/>\n", ElapsedTime);

        // Failed to do MrcSaGvSwitch, HW handshake failed on optimized SaGv Point
        if (MrcStatus != mrcSuccess) {
          return MrcStatus;
        }

        // This SaGv point is done, break out of train loop.
        break;
      }

      for (index = 0; Run && (index < ARRAY_COUNT (MrcCallTable)); index++, post_code++) {

        Task = &MrcCallTable[index];
        //
        // Decide if we need to execute the selected MRC task.
        //
        // If MRC task selected and ExecuteFromSelectPoint == TRUE, it means that we need to execute all MRC tasks from Selected point.
        // By clearing the Select variable we force to continue the execution of all remaining Tasks
        if ((Select != 0) && (ExecuteFromSelectPoint) && (Select == Task->post_code_ovr) ) {
          Select = 0;
        }

        if ((NULL == Task->mrc_task) || (Inputs->Iteration >= Task->iteration)) {
          continue;
        }
        if   (((BootMode == bmS3)   && (Task->policy_flag & MF_S3))
          ||  ((BootMode == bmFast) && (Task->policy_flag & MF_FAST))
          ||  ((BootMode == bmWarm) && (Task->policy_flag & MF_WARM))
          ||  ((BootMode == bmCold) && (Task->policy_flag & MF_COLD))
          ||  ((Select != 0))) {
          if  ((Select == 0) || (Select == Task->post_code_ovr)) {
            if (((MrcIntData->SaGvPoint == SaGvFirst)              && (Task->policy_flag & MF_GV_FIRST)) ||
                ((((SaGvOther >> MrcIntData->SaGvPoint) & 1) != 0) && (Task->policy_flag & MF_GV_OTHER)) ||
                ((MrcIntData->SaGvPoint == SaGvLast)               && (Task->policy_flag & MF_GV_LAST)) ||
                (SaGv != MrcSaGvEnabled)) {
              if (Task->oem_cmd < OemNumOfCommands) {
                CpStatus = MrcInternalCheckPoint (MrcData, Task->oem_cmd, NULL);
                if (mrcSuccess != CpStatus) {
                  continue;
                }
              }
              // Output post code to post code I/O port.
              PostCodeOut = (Task->post_code_ovr == POST_CODE_NO_OVR) ? post_code : Task->post_code_ovr;
              if ((Select == 0) || ((Select > 0) && (Select == PostCodeOut))) {
                MrcCall->MrcDebugHook (MrcData, PostCodeOut);
              }
              // Output debug string to serial output and execute the MRC task.
              MRC_DEBUG_MSG (Debug, MSG_LEVEL_NOTE, "\nMRC task -- %s -- Started.\n", Task->String);
              if (!Task->DebugPrint) {
                DebugLevel = Debug->Level;
                Debug->Level = MSG_LEVEL_NEVER;
              }
              start_time  = MrcCall->MrcGetCpuTime ();
              MrcStatus   = Task->mrc_task (MrcData);
              finish_time = MrcCall->MrcGetCpuTime ();
              if (!Task->DebugPrint) {
                Debug->Level = DebugLevel;
              }
              MRC_DEBUG_MSG (
                Debug,
                MSG_LEVEL_ERROR,
                "MRC task %s -- %s, Status = %Xh.",
                Task->String,
                (mrcSuccess == MrcStatus) ? "SUCCEEDED" : "FAILED",
                MrcStatus
                );
              if (Debug->LowestStackAddr.DataN < LowestStackAddr) {
                MRC_DEBUG_MSG (Debug, MSG_LEVEL_ERROR, " Stack Depth: %d\n", Debug->TopStackAddr.DataN - Debug->LowestStackAddr.DataN);
                LowestStackAddr = Debug->LowestStackAddr.DataN;
              } else {
                MRC_DEBUG_MSG (Debug, MSG_LEVEL_ERROR, "\n");
              }
              if ((mrcSuccess != MrcStatus) || (Select != 0)) {
                Run = 0;  // Stop task execution on failure or running one Task.
              }

              ElapsedTime = (UINT32) (finish_time - start_time);
              TotalTime += ElapsedTime;
              PointTime += ElapsedTime;
              // <MrcTimer Task="Print input parameters" msec="2"/>
              MRC_DEBUG_MSG (Debug, MSG_LEVEL_TIME, "<MrcTimer Task=\"%s\" msec=\"%u\"/>\n", Task->String, ElapsedTime);

              if (Inputs->RmtPerTask && (Task->policy_flag & MF_RMT)) {  // Run RMT after this task
                MrcRankMarginTool (MrcData); // RMT for composite UI error
              }
              if (Inputs->TrainTrace && (Task->policy_flag & MF_TST)) {
                // <MrcTrainedStateTrace> ... <MrcTrainedStateTrace/>
                MRC_DEBUG_MSG (Debug, MSG_LEVEL_NOTE, "<MrcTrainedStateTrace>\n");
                MrcTrainedStateTrace (MrcData);
                MRC_DEBUG_MSG (Debug, MSG_LEVEL_NOTE, "<MrcTrainedStateTrace/>\n");
              }
            }
          } // Select match
        } // if boot mode match
      } // for index
      // Check if MRC failed due to RetrainMarginCheck
      if (MrcStatus == mrcRetrain) {
        MRC_DEBUG_MSG (Debug, MSG_LEVEL_ERROR, "\nRerunning training with higher UPM/PWR limits!\n");

        MrcPrepareNextMrcIteration (MrcData, TRUE);
        MrcIntData->SaGvPoint      = SaGvPoint;
        MRC_DEBUG_MSG_OPEN (MrcData, Inputs->DebugLevel, Inputs->DebugStream);
        Inputs->Iteration++;
        Run = 1;
        RetrainDone = 1;
      } else if (MrcStatus == mrcResetRecovery) {
        if ((Inputs->Iteration++ - RetrainDone) > MRC_MAX_RESET_RECOVERY_ATTEMPTS) {
          MRC_DEBUG_MSG (Debug, MSG_LEVEL_ERROR, "\nReset Recovery Attempts exceeded %d!\n", MRC_MAX_RESET_RECOVERY_ATTEMPTS);
        } else {
          MRC_DEBUG_MSG (Debug, MSG_LEVEL_ERROR, "\nRerunning due to Reset Recovery!\n");
          // Prepare for the next MRC iteration
          MrcPrepareNextMrcIteration (MrcData, FALSE);
          MrcIntData->SaGvPoint = SaGvPoint;
          MRC_DEBUG_MSG_OPEN (MrcData, Inputs->DebugLevel, Inputs->DebugStream);
          Run = 1;
          RetrainLoop--; // RetrainLoop should not change on subsequent loop
        }
      } else {
        break;    // No need to retrain
      }
    } // Retrain loop

    if ((SaGv == MrcSaGvEnabled) && (SaGvPoint < SaGvLast) && (Select == 0)) {
      // Prepare for the next MRC iteration
      MrcPrepareNextMrcIteration (MrcData, FALSE);
    }

    if ((SaGv == MrcSaGvEnabled) && (Select == 0)) {
      MRC_DEBUG_MSG (
        Debug,
        MSG_LEVEL_TIME,
        "MRC timer: Total time for SAGV point %u = %u msec.\n",
        SaGvPoint,
        PointTime
        );
    }
  } // for GvPoint

  if (Select == 0) {
    MRC_DEBUG_MSG (Debug, MSG_LEVEL_TIME, "MRC timer: Total time to execute tasks = %u msec.\n", TotalTime);
  }

  return MrcStatus;
}

/**

@brief

  Get the number of call table entries.

  @param[in, out] MrcData - Include all MRC global data.

  @retval The number of call table entries.

**/
UINT16
MrcGetCallTableCount (
  IN OUT MrcParameters *const MrcData
  )
{
  return ARRAY_COUNT (MrcCallTable);
}
